#pragma once

#include <string>
#include <vector>
#include <regex>
#include <boost/json.hpp>
#include <boost/dynamic_bitset.hpp>
#include <fstream>
#include <sstream>

#include "Util/CpuCoreSelector.hpp"
#include "Util/FileParsing.hpp"
#include "Util/Sharing.hpp"

enum class Analysis {univariate, multivariate, exclusive_multivariate};

class Settings {
  public:
  
  Analysis multivariate_test;

  bool compact_distributions;
  bool minimize_probing_sets;
  bool relaxed_model;
  bool remove_full_probing_sets;
  bool transitional_leakage;
  bool waveform_simulation;
  bool input_is_additive;
  bool output_is_additive;

  uint64_t end_condition_clock_cycles;

  uint64_t input_base;
  uint64_t input_exponent;
  uint64_t output_base;
  uint64_t output_exponent;

  uint64_t maximum_multivariate_distance;
  uint64_t number_of_clock_cycles; 
  uint64_t number_of_entries_in_report;
  uint64_t number_of_faults_per_run;  
  uint64_t number_of_pipeline_stages; // TODO set
  uint64_t number_of_probing_sets_per_step;
  uint64_t number_of_simulations;
  uint64_t number_of_simulations_per_step;
  uint64_t number_of_simulations_per_write; 
  uint64_t number_of_threads;
  uint64_t number_of_wait_cycles;
  uint64_t order_of_test;
  
  double effect_size;
  double alpha_threshold;
  double beta_threshold;

  Polynomial input_irreducible_polynomial;
  Polynomial output_irreducible_polynomial;

  std::string asm_file_path; // settings.externalBinaryInformation.at(2)
  std::string binary_file_path; // settings.externalBinaryInformation.at(0)
  std::string map_file_path; // settings.externalBinaryInformation.at(1)
  std::string clock_signal_name;
  std::string module_name;

  std::vector<uint64_t> faulted_clock_cycles;
  std::vector<uint64_t> test_clock_cycles;

  std::vector<std::vector<TriStateBit>> groups;

  std::vector<std::vector<std::string>> always_random_input_signals;

  std::vector<std::vector<InputAssignment>> input_sequence;

  std::vector<std::pair<std::string, bool>> end_condition_signals;

  std::vector<std::vector<std::string>> output_shares;
  std::vector<std::vector<TriStateBit>> expected_outputs;

  bool IsDistanceSmallEnough(uint64_t distance) const;
  bool IsMultivariateEvaluationRequired() const;

  bool IsConsideredProbeTarget(std::string signal_name) const;
  bool IsConsideredExtension(std::string signal_name) const;

  bool GetHardwareTarget() const;
  TriStateBit GetGroupTriStateBit(uint64_t group_index, uint64_t bit_index) const;
  TriStateBit GetExpectedOutputTriStateBit(uint64_t group_index, uint64_t bit_index) const;
  std::string GetOutputShare(uint64_t share_index, uint64_t bit_index) const;
  uint64_t GetNumberOfEntriesInReport() const;
  uint64_t GetNumberOfPipelineStages() const;
  std::string GetLocationOfCipher() const;
  std::string GetCompilerFlags() const;
  uint64_t GetEndWaitCycles() const;
  uint64_t GetNumberOfSimulations() const;
  uint64_t GetNumberOfSimulationsPerWrite() const;

  /**
   * @brief Constructs a new Settings object with basic settings like max_threads, without design.
   * Used so multithreading can be used for CircuitStruct generation
   * @param config_file_path Relative path to config json file
   * @author Simon Osterheider
  */
  Settings(const std::string& config_file_path);
  
  InputAssignment& GetInputSequence(uint64_t cycle_index, uint64_t input_index){
    return input_sequence[cycle_index][input_index];
  }


private: 

  std::regex considered_probe_targets;
  std::regex considered_extensions;

  bool are_probe_targets_included;
  bool are_extensions_included;

  std::string compiler_flags;  
  uint64_t end_wait_cycles;

  bool hardware_target;
  std::string location_of_cipher;

  void ParseInputFiniteField(const boost::json::object& json_object);
  void ParseOutputFiniteField(const boost::json::object& json_object);
  void ParseAlwaysRandomInputs(const boost::json::object& json_object);
  void ParseClockSignalName(const boost::json::object& json_object);
  void ParseCompactDistributions(const boost::json::object& json_object);
  void ParseCompilerFlags(const boost::json::object& json_object);
  void ParseEffectSize(const boost::json::object& json_object);
  void ParseEndCondition(const boost::json::object& end_condition);
  void ParseEndWaitCycles(const boost::json::object& json_object);
  void ParseConsideredProbeTargets(const boost::json::object& json_object);
  void ParseConsideredExtensions(const boost::json::object& json_object);
  void ParseExpectedOutputs(const boost::json::object& json_object);
  void ParseGroups(const boost::json::object& json_object);
  void ParseHardwareTarget(const boost::json::object& json_object);
  void ParseLocationOfCipher(const boost::json::object& json_object);
  void ParseNumberOfClockCycles(const boost::json::object& json_object);
  void ParseMaximumMultivariateDistance(const boost::json::object& json_object);
  void ParseMinimizeProbingSets(const boost::json::object& json_object);
  void ParseMultivariateTest(const boost::json::object& json_object);
  void ParseNumberOfEntriesInReport(const boost::json::object& json_object);
  void ParseNumberOfPipelineStages(const boost::json::object& json_object);
  void ParseNumberOfProbingSetsPerStep(const boost::json::object& json_object);
  void ParseNumberOfSimulations(const boost::json::object& json_object);
  void ParseNumberOfSimulationsPerStep(const boost::json::object& json_object);
  void ParseNumberOfSimulationsPerWrite(const boost::json::object& json_object);
  void ParseNumberOfThreads(const boost::json::object& json_object);
  void ParseOrderOfTest(const boost::json::object& json_object);
  void ParseOutputShares(const boost::json::object& json_object);
  void ParseRemoveFullProbingSets(const boost::json::object& json_object);
  void ParseTestClockCycles(const boost::json::object& json_object);
  void ParseTransitionalLeakage(const boost::json::object& json_object);
  void ParseWaveformSimulation(const boost::json::object& json_object);
  void ParseInputSequence(const boost::json::object& json_object);
  void ParseRelaxedModel(const boost::json::object& json_object);
  void ParseAlphaThreshold(const boost::json::object& json_object);
  void ParseBetaThreshold(const boost::json::object& json_object);
};

namespace SettingNames {
  extern const std::string CLOCK_SIGNAL_NAME;
  extern const std::string NUMBER_OF_THREADS;
  extern const std::string COMPACT_DISTRIBUTIONS;
  extern const std::string EFFECT_SIZE;
  extern const std::string END_CONDITION;
  extern const std::string END_WAIT_CYCLES;
  extern const std::string EXPECTED_OUTPUT;
  extern const std::string NUMBER_OF_CLOCK_CYCLES;
  extern const std::string MAXIMUM_MULTIVARIATE_DISTANCE;
  extern const std::string MINIMIZE_PROBING_SETS;
  extern const std::string MULTIVARIATE_TEST;
  extern const std::string ALWAYS_RANDOM_INPUTS;
  extern const std::string NUMBER_OF_ENTRIES_IN_REPORT;
  extern const std::string NUMBER_OF_INITIAL_INPUTS;
  extern const std::string GROUPS;
  extern const std::string INPUT_SEQUENCE;
  extern const std::string HARDWARE_TARGET;
  extern const std::string OUTPUT_SHARES;
  extern const std::string NUMBER_OF_PROBING_SETS_PER_STEP;
  extern const std::string NUMBER_OF_SIMULATIONS;
  extern const std::string NUMBER_OF_SIMULATIONS_PER_STEP;
  extern const std::string NUMBER_OF_SIMULATIONS_PER_WRITE;
  extern const std::string TEST_CLOCK_CYCLES;
  extern const std::string ORDER_OF_TEST;
  extern const std::string CONSIDERED_PROBE_TARGETS;
  extern const std::string CONSIDERED_EXTENSIONS;
  extern const std::string REMOVE_FULL_PROBING_SETS;
  extern const std::string TRANSITIONAL_LEAKAGE;
  extern const std::string WAVEFORM_SIMULATION;
  extern const std::string COMPILER_FLAGS;
  extern const std::string LOCATION_OF_CIPHER;
  extern const std::string NUMBER_OF_PIPELINE_STAGES;
  extern const std::string RELAXED_MODEL;
  extern const std::string ALPHA_THRESHOLD;
  extern const std::string BETA_THRESHOLD; 
  extern const std::string INPUT_FINITE_FIELD;
  extern const std::string OUTPUT_FINITE_FIELD; 
} // SettingNames