#include "Hardware/Definitions.hpp"

bool Hardware::CircuitStruct::IsProbeOnSignalAllowed(int base_signal_index,
                                                     int current_signal_index,
                                                     int clock_signal_index) {
  return !Signals[current_signal_index]->Deleted &&
         (current_signal_index != clock_signal_index) &&
         ((Signals[current_signal_index]->ProbeAllowed) ||
          (Signals[base_signal_index]->ProbeAllowed == 2));
}

bool Hardware::CircuitStruct::IsFaultOnSignalAllowed(int signal_index,
                                                     int clock_signal_index) {
  return !Signals[signal_index]->Deleted &&
         (signal_index != clock_signal_index) &&
         ((Signals[signal_index]->FaultAllowed) ||
          (Signals[signal_index]->FaultAllowed == 2));
}

bool Hardware::CircuitStruct::CanProbeOnSignalBePropagated(int signal_index,
                                                           Library& library) {
  return (Signals[signal_index]->Output != -1) &&
         (!library.IsCellRegister(Cells[Signals[signal_index]->Output]->Type));
}

bool Hardware::CircuitStruct::IsGateThatOutputsSignalDeleted(int signal_index) {
  return Cells[Signals[signal_index]->Output]->Deleted == 1;
}

int Hardware::CircuitStruct::GetNumberOfInputsForSignalsComputingCell(
    int signal_index) {
  if (Signals[signal_index]->Output == -1) {
    std::string error_message =
        "Tried to access a cell that computes \"" +
        (std::string)Signals[signal_index]->Name + "\". As  \"" +
        (std::string)Signals[signal_index]->Name +
        "\" is a primary input, such a cell cannot be found!";
    throw std::logic_error(error_message);
  } else {
    return Cells[Signals[signal_index]->Output]->NumberOfInputs;
  }
}

std::vector<uint64_t> Hardware::CircuitStruct::GetSignals() {
  std::vector<uint64_t> result;
  int cell_index, output_index;

  result.push_back(0); // Add 1'b0 and 1'b1
  result.push_back(1); // TODO: Change

  for (cell_index = 0; cell_index < NumberOfInputs; ++cell_index) {
    result.push_back(Inputs[cell_index]);
  }

  for (cell_index = 0; cell_index < NumberOfRegs; ++cell_index) {
    for (output_index = 0;
         output_index < Cells[Regs[cell_index]]->NumberOfOutputs;
         ++output_index) {
      if (Cells[Regs[cell_index]]->Outputs[output_index] != -1) {
        result.push_back(Cells[Regs[cell_index]]->Outputs[output_index]);
      }
    }
  }

  for (short depth_index = 1; depth_index <= MaxDepth; ++depth_index) {
    for (cell_index = 0; cell_index < NumberOfCellsInDepth[depth_index];
         ++cell_index) {
      for (output_index = 0;
           output_index <
           Cells[CellsInDepth[depth_index][cell_index]]->NumberOfOutputs;
           ++output_index) {
        result.push_back(Cells[CellsInDepth[depth_index][cell_index]]
                             ->Outputs[output_index]);
      }
    }
  }

  return result;
}

uint64_t Hardware::CircuitStruct::GetSignalIndexByName(const std::string& signal_name){
  uint64_t index, number_of_signals = NumberOfSignals;
  std::string current_signal_name;

  for (index = 0; index < number_of_signals; ++index) {
    current_signal_name = Signals[index]->Name;

    if (current_signal_name == signal_name) {
      return index;
    }
  }

  if (index == number_of_signals) {
    throw std::logic_error("Error while searching a signal. The signal \"" + signal_name + "\" could not be found!");
  }

  return 0;
}

Simulation::Simulation(Hardware::CircuitStruct& circuit, Settings& settings){
  uint64_t index, signal_index, signal_value;
  std::string fresh_mask_signal_names;

	number_of_clock_cycles_ = settings.number_of_clock_cycles;
	if ((settings.end_condition_clock_cycles) &&
		((settings.end_condition_clock_cycles + 1 + settings.number_of_wait_cycles) < settings.end_condition_clock_cycles))
		number_of_clock_cycles_ = settings.end_condition_clock_cycles + 1 + settings.number_of_wait_cycles;

    number_of_processed_simulations = 0;
    selected_groups_.resize(settings.number_of_simulations_per_step);

    clock_signal_index_ = circuit.GetSignalIndexByName(settings.clock_signal_name);

    always_random_inputs_indices_.resize(settings.always_random_input_signals.size());
    for (index = 0; index < settings.always_random_input_signals.size(); ++index) {
      fresh_mask_signal_names += "{";
      for (std::string& signal_name : settings.always_random_input_signals[index]) {
        signal_index = circuit.GetSignalIndexByName(signal_name);
        always_random_inputs_indices_[index].push_back(signal_index);
        fresh_mask_signal_names += signal_name + ", ";
      }

      if (fresh_mask_signal_names.length() > 2) {
        fresh_mask_signal_names.erase(fresh_mask_signal_names.length() - 2);
      }

      fresh_mask_signal_names += "}, ";
    }

    if (fresh_mask_signal_names.length() > 2) {
      fresh_mask_signal_names.erase(fresh_mask_signal_names.length() - 2);
    }

    std::cout << "Successfully matched " << always_random_inputs_indices_.size() << " fresh mask signals [" << fresh_mask_signal_names << "]." << std::endl;

    for (std::pair<std::string, bool>& end_condition_signal : settings.end_condition_signals) {
      signal_index = circuit.GetSignalIndexByName(end_condition_signal.first);
      signal_value = end_condition_signal.second ? 0xffffffffffffffff : 0x0000000000000000;
      end_condition_signals_.emplace_back(signal_index, signal_value);
    }

    if (settings.output_shares.size()){
      output_share_signal_indices_.resize(settings.output_shares.size());

      uint64_t number_of_values, number_of_bits;

      for (uint64_t share_index = 0; share_index < settings.output_shares.size(); ++share_index) {
          number_of_bits = std::ceil(std::log2l(settings.output_base)) * settings.output_exponent;
          number_of_values = settings.output_shares[share_index].size() / number_of_bits;
          output_share_signal_indices_[share_index].resize(number_of_values);

          for (uint64_t value_index = 0; value_index < number_of_values; ++value_index) {
              for (uint64_t bit_index = 0; bit_index < number_of_bits; ++bit_index) {
                  output_share_signal_indices_[share_index][value_index].push_back(circuit.GetSignalIndexByName(settings.output_shares[share_index][value_index * number_of_bits + bit_index])); 
              }
          }
      }

      expected_unshared_output_values_.resize(settings.groups.size());
      for (uint64_t share_index = 0; share_index < settings.groups.size(); ++share_index) {
          number_of_bits = std::ceil(std::log2l(settings.output_base)) * settings.output_exponent;
          number_of_values = settings.output_shares[share_index].size() / number_of_bits;
          expected_unshared_output_values_[share_index].resize(number_of_values);

          for (uint64_t value_index = 0; value_index < number_of_values; ++value_index) {
              for (uint64_t bit_index = 0; bit_index < number_of_bits; ++bit_index) {
                  expected_unshared_output_values_[share_index][value_index].push_back(settings.expected_outputs[share_index][value_index * number_of_bits + bit_index]);
              }
          }
      }
    }
}