#include "Hardware/Propagation.hpp"

namespace Hardware{
    template <class ExtensionContainer>
    void RemoveDuplicates(std::vector<ExtensionContainer>& extensions) {
        std::sort(extensions.begin(), extensions.end(), std::greater<ExtensionContainer>());
        extensions.erase(std::unique(extensions.begin(), extensions.end()), extensions.end());
    }

    template void RemoveDuplicates(std::vector<RobustProbe>&);

    template <class ExtensionContainer>
    Propagation<ExtensionContainer>::Propagation() {}
    template Propagation<RobustProbe>::Propagation();
    template Propagation<RelaxedProbe>::Propagation();

    template <class ExtensionContainer>
    Propagation<ExtensionContainer>::Propagation(uint64_t signal_index) : signal_index_(signal_index) {}
    template Propagation<RobustProbe>::Propagation(uint64_t);
    template Propagation<RelaxedProbe>::Propagation(uint64_t);

    template <class ExtensionContainer>
    Propagation<ExtensionContainer>::Propagation(uint64_t signal_index, std::vector<ExtensionContainer> extension_indies) : signal_index_(signal_index), extension_indices_(extension_indies) {}
    
    template Propagation<RobustProbe>::Propagation(uint64_t, std::vector<RobustProbe>);
    template Propagation<RelaxedProbe>::Propagation(uint64_t, std::vector<RelaxedProbe>);
    
    template <class ExtensionContainer>
    Propagation<ExtensionContainer>::Propagation(uint64_t signal_index, Library& library, CircuitStruct& circuit, uint64_t clock_signal_index, std::vector<Propagation>& propagations) : signal_index_(signal_index) {
        Propagate(library, circuit, clock_signal_index, propagations, signal_index);
    }

    template Propagation<RobustProbe>::Propagation(uint64_t, Library&, CircuitStruct&, uint64_t, std::vector<Propagation>&);
    template Propagation<RelaxedProbe>::Propagation(uint64_t, Library&, CircuitStruct&, uint64_t, std::vector<Propagation>&);

    template <>
    Propagation<RelaxedProbe>::Propagation(uint64_t signal_index, uint64_t enable_index, uint64_t number_of_enable_indices, uint64_t number_of_signal_indices, std::vector<uint64_t> signal_indices, std::vector<uint64_t> propagation_indices) : signal_index_(signal_index) {
        extension_indices_.resize(1);
        extension_indices_[0].enable_index_ = enable_index;
        extension_indices_[0].number_of_enable_indices_ = number_of_enable_indices;
        extension_indices_[0].number_of_signal_indices_ = number_of_signal_indices;
        extension_indices_[0].signal_indices_ = signal_indices;
        extension_indices_[0].propagation_indices_ = propagation_indices;
    }

    template <>
    bool Propagation<RobustProbe>::operator==(const Propagation<RobustProbe>& other) const {
        return (signal_index_ == other.signal_index_) && (extension_indices_ == other.extension_indices_);
    }

    template <>
    bool Propagation<RelaxedProbe>::operator==(const Propagation<RelaxedProbe>& other) const {
        return (signal_index_ == other.signal_index_)  && (extension_indices_[0].enable_index_ == other.extension_indices_[0].enable_index_) &&
        (extension_indices_[0].number_of_enable_indices_ == other.extension_indices_[0].number_of_enable_indices_) && (extension_indices_[0].number_of_signal_indices_ == other.extension_indices_[0].number_of_signal_indices_) &&
        (extension_indices_[0].propagation_indices_ == other.extension_indices_[0].propagation_indices_) && (extension_indices_[0].signal_indices_ == other.extension_indices_[0].signal_indices_);
    }

    template <class ExtensionContainer>
    uint64_t Propagation<ExtensionContainer>::GetSignalIndex() {
        return signal_index_;
    }

    template uint64_t Propagation<RobustProbe>::GetSignalIndex();
    template uint64_t Propagation<RelaxedProbe>::GetSignalIndex();  

    template <>
    std::vector<uint64_t> Propagation<RobustProbe>::GetExtensionIndices() {
        return extension_indices_;
    }    

    template <>
    std::vector<uint64_t> Propagation<RelaxedProbe>::GetExtensionIndices(std::vector<Propagation>& propagations) {
        std::vector<uint64_t> result, indices, signal_indices;
        std::set<uint64_t> considered_indices;
        uint64_t index;

        indices.push_back(signal_index_);

        while(!indices.empty()){
            index = indices.back();
            indices.pop_back();
            
            std::vector<Propagation>::iterator it = std::lower_bound(propagations.begin(), propagations.end(), index, [](Propagation& propagation, uint64_t value) {
                return propagation.GetSignalIndex() < value;
            });

            if (it != propagations.end() && index == it->GetSignalIndex()) {
                index = std::distance(propagations.begin(), it);
            }else{
                throw std::out_of_range("Unable to find the propagation belonging to a signal! It seems that you found a bug in PROLEAD. Please get in touch with me (nicolai.mueller@rub.de) so that we can fix this issue!");
            }

            signal_indices = propagations[index].extension_indices_[0].signal_indices_;
            result.insert(result.end(), signal_indices.begin(), signal_indices.end());

            for (size_t propagation_index : propagations[index].extension_indices_[0].propagation_indices_){
                if (considered_indices.find(propagation_index) == considered_indices.end()){
                    indices.push_back(propagation_index);
                    considered_indices.insert(propagation_index);
                }
            }
        }

        std::sort(result.begin(), result.end());
        result.erase(std::unique(result.begin(), result.end()), result.end());
        return result;
    }

    template <>
    std::vector<uint64_t> Propagation<RelaxedProbe>::GetEnableIndices(std::vector<Propagation>& propagations) {
        std::vector<uint64_t> result, indices, signal_indices;
        std::set<uint64_t> considered_indices;
        uint64_t index;

        if (extension_indices_[0].number_of_enable_indices_){
            indices.push_back(signal_index_);
            result.push_back(extension_indices_[0].enable_index_);
        }

        while(!indices.empty()){
            index = indices.back();
            indices.pop_back();
            
            std::vector<Propagation>::iterator it = std::lower_bound(propagations.begin(), propagations.end(), index, [](Propagation& propagation, uint64_t value) {
                return propagation.GetSignalIndex() < value;
            });

            if (it != propagations.end() && index == it->GetSignalIndex()) {
                index = std::distance(propagations.begin(), it);
            }else{
                throw std::out_of_range("Unable to find the propagation belonging to a signal! It seems that you found a bug in PROLEAD. Please get in touch with me (nicolai.mueller@rub.de) so that we can fix this issue!");
            }

            if (propagations[index].extension_indices_[0].number_of_enable_indices_){
                if (std::find(result.begin(), result.end(), propagations[index].extension_indices_[0].enable_index_) == result.end()){
                    result.push_back(propagations[index].extension_indices_[0].enable_index_);
                }
            }

            for (size_t propagation_index : propagations[index].extension_indices_[0].propagation_indices_){
                if (considered_indices.find(propagation_index) == considered_indices.end()){
                    indices.push_back(propagation_index);
                    considered_indices.insert(propagation_index);
                }
            }
        }

        return result;
    }

    template <>
    uint64_t Propagation<RelaxedProbe>::GetEnableIndex() {
        return extension_indices_[0].enable_index_;
    }

    template <>
    std::vector<RobustProbe>& Propagation<RobustProbe>::GetExtensionContainer() {
        return extension_indices_;
    }

    template <>
    uint64_t Propagation<RelaxedProbe>::GetNumberOfEnableIndices(){
        return extension_indices_[0].number_of_enable_indices_;
    }

    template <>
    uint64_t Propagation<RelaxedProbe>::GetNumberOfSignalIndices(){
        return extension_indices_[0].number_of_signal_indices_;
    }

    template <>
    RelaxedProbe* Propagation<RelaxedProbe>::GetProbeAddress(uint64_t index){
        return &extension_indices_[index];
    }

    template <> 
    uint64_t Propagation<RelaxedProbe>::BackpropagateUntilBranch(CircuitStruct& circuit, uint64_t clock_signal_index, uint64_t base_index, uint64_t signal_index){
        uint64_t result = signal_index;

        while ((circuit.Signals[result]->Output != -1) && (circuit.GetNumberOfInputsForSignalsComputingCell(result) == 1) && circuit.IsProbeOnSignalAllowed(base_index, result, clock_signal_index)){
            result = circuit.Cells[circuit.Signals[result]->Output]->Inputs[0];
        }   

        return result;     
    }

    template <>
    void Propagation<RobustProbe>::Propagate(Library& library, CircuitStruct& circuit, uint64_t clock_signal_index, std::vector<Propagation>& propagations, uint64_t base_index) {
        int index, input_index;

        if (circuit.IsProbeOnSignalAllowed(base_index, signal_index_, clock_signal_index)) {
            if (circuit.CanProbeOnSignalBePropagated(signal_index_, library)){
                for (index = 0; index < circuit.GetNumberOfInputsForSignalsComputingCell(signal_index_); ++index) {
                    input_index = circuit.Cells[circuit.Signals[signal_index_]->Output]->Inputs[index];
                    
                    if (circuit.IsProbeOnSignalAllowed(base_index, input_index, clock_signal_index) && propagations[input_index].extension_indices_.empty()){
                        propagations[input_index].Propagate(library, circuit, clock_signal_index, propagations, base_index);
                    }

                    extension_indices_.insert(extension_indices_.end(), propagations[input_index].extension_indices_.begin(), propagations[input_index].extension_indices_.end());
                }

                std::sort(extension_indices_.begin(), extension_indices_.end());
                extension_indices_.erase(std::unique(extension_indices_.begin(), extension_indices_.end()), extension_indices_.end());
            }else{
                extension_indices_.push_back(signal_index_);
            }
        }   
    }

    template <>
    void Propagation<RelaxedProbe>::Propagate(Library& library, CircuitStruct& circuit, uint64_t clock_signal_index, std::vector<Propagation>& propagations, uint64_t base_index) {        
        int index, input_index;
        uint64_t signal_index = BackpropagateUntilBranch(circuit, clock_signal_index, base_index, signal_index_);

        extension_indices_.resize(1);
        extension_indices_[0].number_of_enable_indices_ = 0;
        extension_indices_[0].number_of_signal_indices_ = 0;
        if (circuit.IsProbeOnSignalAllowed(base_index, signal_index, clock_signal_index)) {
            if (circuit.CanProbeOnSignalBePropagated(signal_index, library)){
                extension_indices_[0].enable_index_ = signal_index;
                ++extension_indices_[0].number_of_enable_indices_;

                for (index = 0; index < circuit.GetNumberOfInputsForSignalsComputingCell(signal_index); ++index) {
                    input_index = circuit.Cells[circuit.Signals[signal_index]->Output]->Inputs[index];
                    input_index = BackpropagateUntilBranch(circuit, clock_signal_index, base_index, input_index);

                    if (circuit.IsProbeOnSignalAllowed(base_index, input_index, clock_signal_index) && propagations[input_index].extension_indices_[0].signal_indices_.empty()){
                        propagations[input_index].Propagate(library, circuit, clock_signal_index, propagations, base_index);
                    }

                    if (!propagations[input_index].extension_indices_[0].signal_indices_.empty()){
                        extension_indices_[0].number_of_enable_indices_ += propagations[input_index].extension_indices_[0].number_of_enable_indices_;
                        extension_indices_[0].number_of_signal_indices_ += propagations[input_index].extension_indices_[0].number_of_signal_indices_;
                        extension_indices_[0].propagation_indices_.push_back(input_index);
                    }
                }
            }

            extension_indices_[0].signal_indices_.push_back(signal_index);

            if (extension_indices_[0].number_of_signal_indices_ == 0){
                ++extension_indices_[0].number_of_signal_indices_;
            } 
        }
    }

    template <>
    bool Propagation<RobustProbe>::IsObsolete(Library& library, CircuitStruct& circuit){
        uint64_t signal_index = GetSignalIndex();

        if (circuit.Signals[signal_index]->ProbeAllowed != 2) {
            return true;
        }

        if ((GetExtensionIndices().size() == 0) || (signal_index < (uint64_t)circuit.NumberOfConstants)){
            return true;
        }
         
        for (int index = 0; index < circuit.Signals[signal_index]->NumberOfInputs; ++index) {
            if (!library.IsCellRegister(circuit.Cells[circuit.Signals[signal_index]->Inputs[index]]->Type)){
                return true;
            }
        }

        return false;
    }   

    template <>
    bool Propagation<RelaxedProbe>::IsObsolete(Library&, CircuitStruct& circuit){
        if (extension_indices_.empty()){
            return true;
        }

        if (!extension_indices_[0].number_of_enable_indices_ && extension_indices_[0].propagation_indices_.empty() && !extension_indices_[0].number_of_signal_indices_){
            return true;
        }

        if (circuit.Signals[GetSignalIndex()]->Output != -1){
            if (circuit.Cells[circuit.Signals[GetSignalIndex()]->Output]->NumberOfInputs == 1){
                return true;
            }
        }

        return false;
    }  

    template <>
    Propagation<RelaxedProbe> Propagation<RelaxedProbe>::ExtendWithTime(CircuitStruct&, uint64_t clock_cycle, std::vector<Probe>& probes, std::vector<Enabler<CustomOperation>>& enabler) {        
        Propagation<RelaxedProbe> propagation;
        std::vector<uint64_t> indices;
        uint64_t index;
        propagation.signal_index_ = SearchProbe(signal_index_, clock_cycle, probes);
        propagation.extension_indices_.resize(1);

        if (clock_cycle){
            propagation.extension_indices_[0].signal_indices_.resize(2);
            propagation.extension_indices_[0].signal_indices_[0] = SearchProbe(extension_indices_[0].signal_indices_[0], clock_cycle - 1, probes);
            propagation.extension_indices_[0].signal_indices_[1] = SearchProbe(extension_indices_[0].signal_indices_[0], clock_cycle, probes);
        }else{
            propagation.extension_indices_[0].signal_indices_.resize(1);
            propagation.extension_indices_[0].signal_indices_[0] = SearchProbe(extension_indices_[0].signal_indices_[0], clock_cycle, probes);
        }

        if (extension_indices_[0].number_of_enable_indices_){
            index = SearchProbe(extension_indices_[0].enable_index_, clock_cycle, probes);
            propagation.extension_indices_[0].enable_index_ = SearchEnabler(index, enabler);
        }

        propagation.extension_indices_[0].number_of_enable_indices_ = extension_indices_[0].number_of_enable_indices_;

        for (uint64_t propagation_index : extension_indices_[0].propagation_indices_){
            indices.push_back(SearchProbe(propagation_index, clock_cycle, probes));
        }

        propagation.extension_indices_[0].propagation_indices_ = indices;

        return propagation;
    }

    template <>
    void Propagation<RelaxedProbe>::Finalize(std::vector<Propagation<RelaxedProbe>>& propagations) {    
        std::vector<Propagation<RelaxedProbe>>::iterator it;

        for (uint64_t& index : extension_indices_[0].propagation_indices_){
            it = std::lower_bound(propagations.begin(), propagations.end(), index, [](Propagation<RelaxedProbe>& p, uint64_t signal_index) {
                return p.GetSignalIndex() < signal_index;
            });

            if (it != propagations.end() && it->GetSignalIndex() == index) {
                index = std::distance(propagations.begin(), it);
            } else {
                throw std::out_of_range("Unable to find the propagation belonging to a signal! It seems that you found a bug in PROLEAD. Please get in touch with me (nicolai.mueller@rub.de) so that we can fix this issue!");
            }
        }
    }

    template <>
    void Propagation<RelaxedProbe>::UpdateNumberOfSignals(std::vector<Propagation<RelaxedProbe>>& propagations) { 
        extension_indices_[0].number_of_signal_indices_ = GetExtensionIndices(propagations).size();
        extension_indices_[0].number_of_enable_indices_ = GetEnableIndices(propagations).size();
    } 
}