#include "Util/Settings.hpp"

bool Settings::IsMultivariateEvaluationRequired() const {
  return (multivariate_test != Analysis::univariate) &&
         (test_clock_cycles.size() > 1) && (order_of_test != 1);
}

bool Settings::IsDistanceSmallEnough(uint64_t distance) const {
  bool is_distance_small_enough = distance <= maximum_multivariate_distance;
  bool is_no_unallowed_univariate_set =
      (multivariate_test != Analysis::exclusive_multivariate) || distance != 0;
  return is_distance_small_enough && is_no_unallowed_univariate_set;
}

bool Settings::GetHardwareTarget() const { return number_of_threads; }

TriStateBit Settings::GetGroupTriStateBit(uint64_t group_index,
                                          uint64_t bit_index) const {
  return groups[group_index][bit_index];
}

TriStateBit Settings::GetExpectedOutputTriStateBit(uint64_t share_index,
                                                   uint64_t bit_index) const {
  return expected_outputs[share_index][bit_index];
}

std::string Settings::GetOutputShare(uint64_t group_index,
                                     uint64_t bit_index) const {
  return output_shares[group_index][bit_index];
}

uint64_t Settings::GetNumberOfEntriesInReport() const {
  return number_of_entries_in_report;
}

uint64_t Settings::GetNumberOfPipelineStages() const {
  return number_of_pipeline_stages;
}

std::string Settings::GetLocationOfCipher() const { return location_of_cipher; }

std::string Settings::GetCompilerFlags() const { return compiler_flags; }

uint64_t Settings::GetEndWaitCycles() const { return end_wait_cycles; }

uint64_t Settings::GetNumberOfSimulations() const {
  return number_of_simulations;
}

uint64_t Settings::GetNumberOfSimulationsPerWrite() const {
  return number_of_simulations_per_write;
}

bool Settings::IsConsideredProbeTarget(std::string signal_name) const {
  if (are_probe_targets_included) {
    return std::regex_match(signal_name, considered_probe_targets);
  } else {
    return !std::regex_match(signal_name, considered_probe_targets);
  }
}

bool Settings::IsConsideredExtension(std::string signal_name) const {
  if (are_extensions_included) {
    return std::regex_match(signal_name, considered_extensions);
  } else {
    return !std::regex_match(signal_name, considered_extensions);
  }
}

void Settings::ParseAlwaysRandomInputs(const boost::json::object& json_object) {
  SignalNameGrammar grammar;
  std::vector<std::string> signal_names, parsed;
  std::string signal_name;
  boost::json::array json_array;
  std::string error_context = "Error while parsing the always random inputs: ";
  uint64_t element_size = std::ceil(std::log2l(input_base)) * input_exponent;
  uint64_t number_of_elements;

  if (json_object.contains(SettingNames::ALWAYS_RANDOM_INPUTS) &&
      !hardware_target) {
    throw std::invalid_argument(
        error_context + "This setting is not supported in the software case.");
  }

  SetValue(json_object, SettingNames::ALWAYS_RANDOM_INPUTS, json_array);

  for (auto& input_signal : json_array) {
    if (!input_signal.is_string()) {
      throw std::invalid_argument(
          error_context + "The fresh mask inputs must be given as strings.");
    } else {
      signal_name = input_signal.as_string().c_str();
      parsed = grammar.Parse(signal_name);
      signal_names.insert(signal_names.end(), parsed.begin(), parsed.end());

      if (signal_names.size() % element_size) {
        throw std::invalid_argument(
            error_context +
            "The fresh mask inputs must be a multiple of the element size "
            "defined with the input finite field.");
      }
    }
  }

  number_of_elements = signal_names.size() / element_size;
  always_random_input_signals.resize(number_of_elements);

  for (uint64_t index = 0; index < number_of_elements; ++index) {
    always_random_input_signals[index] = std::vector<std::string>(
        signal_names.begin() + index * element_size,
        signal_names.begin() + (index + 1) * element_size);
  }
}

void Settings::ParseClockSignalName(const boost::json::object& json_object) {
  if (json_object.contains(SettingNames::CLOCK_SIGNAL_NAME)) {
    if (hardware_target) {
      SetValue(json_object, SettingNames::CLOCK_SIGNAL_NAME, clock_signal_name);
    } else {
      throw std::runtime_error(
          "Error while parsing the clock signal name: This settings is not "
          "supported in the software case.");
    }
  } else {
    if (hardware_target) {
      throw std::runtime_error(
          "Error while parsing the clock signal name: The clock signal name is "
          "missing.");
    }
  }
}

void Settings::ParseCompactDistributions(
    const boost::json::object& json_object) {
  compact_distributions = false;
  SetValue(json_object, SettingNames::COMPACT_DISTRIBUTIONS,
           compact_distributions);
}

void Settings::ParseCompilerFlags(const boost::json::object& json_object) {
  compiler_flags = "";
  if (json_object.contains(SettingNames::COMPILER_FLAGS)) {
    SetValue(json_object, SettingNames::COMPILER_FLAGS, compiler_flags);
  } else {
    if (!hardware_target) {
      throw std::runtime_error(
          "Error while parsing the compiler flags: The compiler flags are "
          "missing.");
    }
  }
}

void Settings::ParseEffectSize(const boost::json::object& json_object) {
  effect_size = 0.1;
  SetValue(json_object, SettingNames::EFFECT_SIZE, effect_size);

  if ((effect_size < 0) || (effect_size > 1)) {
    throw std::out_of_range(
        "Error while parsing the effect size: The effect size must be "
        "between "
        "0 and 1.");
  }
}

void Settings::ParseAlphaThreshold(const boost::json::object& json_object) {
  alpha_threshold = 5.0;
  SetValue(json_object, SettingNames::ALPHA_THRESHOLD, alpha_threshold);

  if ((alpha_threshold < 0)) {
    throw std::out_of_range(
        "Error while parsing the alpha threshold: The alpha threshold must be "
        "larger than 0!");
  }
}

void Settings::ParseBetaThreshold(const boost::json::object& json_object) {
  beta_threshold = 0.00001;
  SetValue(json_object, SettingNames::BETA_THRESHOLD, beta_threshold);

  if ((beta_threshold < 0)) {
    throw std::out_of_range(
        "Error while parsing the beta threshold: The beta threshold must be "
        "larger than 0!");
  }
}

void Settings::ParseInputFiniteField(const boost::json::object& json_object) {
  input_base = 2;
  input_exponent = 1;
  input_is_additive = true;
  input_irreducible_polynomial.clear();
  boost::json::object input_finite_field;

  SetValue(json_object, SettingNames::INPUT_FINITE_FIELD, input_finite_field);

  if (json_object.contains(SettingNames::INPUT_FINITE_FIELD)) {
    if (input_finite_field.contains("base")) {
      SetValue(input_finite_field, "base", input_base);
    } else {
      throw std::runtime_error(
          "Error while parsing the input finite field: The base is missing.");
    }

    if (input_finite_field.contains("exponent")) {
      SetValue(input_finite_field, "exponent", input_exponent);
    } else {
      throw std::runtime_error(
          "Error while parsing the input finite field: The exponent is "
          "missing.");
    }

    if (input_finite_field.contains("irreducible_polynomial")) {
      std::string polynomial_string;
      SetValue(input_finite_field, "irreducible_polynomial", polynomial_string);
      IrreduciblePolynomialGrammar grammar;
      input_irreducible_polynomial =
          grammar.Parse(polynomial_string, input_base, input_exponent);
    }

    if (input_finite_field.contains("is_additive")) {
      SetValue(input_finite_field, "is_additive", input_is_additive);
    }
  }
}

void Settings::ParseOutputFiniteField(const boost::json::object& json_object) {
  output_base = 2;
  output_exponent = 1;
  output_is_additive = true;
  output_irreducible_polynomial.clear();
  boost::json::object output_finite_field;

  SetValue(json_object, SettingNames::OUTPUT_FINITE_FIELD, output_finite_field);

  if (json_object.contains(SettingNames::OUTPUT_FINITE_FIELD)) {
    if (output_finite_field.contains("base")) {
      SetValue(output_finite_field, "base", output_base);
    } else {
      throw std::runtime_error(
          "Error while parsing the output finite field: The base is missing.");
    }

    if (output_finite_field.contains("exponent")) {
      SetValue(output_finite_field, "exponent", output_exponent);
    } else {
      throw std::runtime_error(
          "Error while parsing the output finite field: The exponent is "
          "missing.");
    }

    if (output_finite_field.contains("irreducible_polynomial")) {
      std::string polynomial_string;
      SetValue(output_finite_field, "irreducible_polynomial",
               polynomial_string);
      IrreduciblePolynomialGrammar grammar;
      output_irreducible_polynomial =
          grammar.Parse(polynomial_string, output_base, output_exponent);
    }

    if (output_finite_field.contains("is_additive")) {
      SetValue(output_finite_field, "is_additive", output_is_additive);
    }
  }
}

void Settings::ParseEndCondition(const boost::json::object& json_object) {
  boost::json::object end_condition;
  std::string error_context = "Error while parsing the end condition: ";
  SetValue(json_object, SettingNames::END_CONDITION, end_condition);

  if (json_object.contains(SettingNames::END_CONDITION)) {
    if (end_condition.contains("clock_cycles") &&
        end_condition.contains("signals")) {
      throw std::invalid_argument(
          error_context +
          "The end condition must be either depend on the number of clock "
          "cycles or on specific signals!");
    } else if (end_condition.contains("clock_cycles") &&
               !end_condition.contains("signals")) {
      SetValue(end_condition, "clock_cycles", end_condition_clock_cycles);
    } else if (!end_condition.contains("clock_cycles") &&
               end_condition.contains("signals")) {
      boost::json::array signals;
      SetValue(end_condition, "signals", signals);

      for (auto& signal : signals) {
        if (!signal.is_object()) {
          throw std::invalid_argument(
              error_context + "The signals are not specified as json objects!");
        } else if (!signal.as_object().contains("name") ||
                   !signal.as_object().contains("value")) {
          throw std::invalid_argument(
              error_context +
              "The signals do not contain the keys 'name' and 'value'!");
        } else {
          std::string signal_name, signal_value;
          SignalNameGrammar name_grammar;
          std::vector<TriStateBit> values;
          std::vector<std::string> names;

          SetValue(signal.as_object(), "name", signal_name);
          SetValue(signal.as_object(), "value", signal_value);
          names = name_grammar.Parse(signal_name);

          VerilogBitstringGrammar value_grammar(signal_value);
          values = value_grammar.Parse(signal_value);

          for (uint64_t index = 0; index < names.size(); ++index) {
            if (values[index] == TriStateBit::zero_value) {
              end_condition_signals.push_back(
                  std::make_pair(names[index], false));
            } else if (values[index] == TriStateBit::one_value) {
              end_condition_signals.push_back(
                  std::make_pair(names[index], true));
            } else {
              throw std::invalid_argument(
                  error_context +
                  "The signal value is invalid, must be zero or one!");
            }
          }
        }
      }
    }
  }
}

void Settings::ParseEndWaitCycles(const boost::json::object& json_object) {
  end_wait_cycles = 0;
  SetValue(json_object, SettingNames::END_WAIT_CYCLES, end_wait_cycles);
}

void Settings::ParseConsideredProbeTargets(
    const boost::json::object& json_object) {
  std::string action, regex_pattern = ".*";
  boost::json::object considered_probe_targets_object;
  std::string error_context =
      "Error while parsing the considered probe targets: ";
  are_probe_targets_included = true;

  SetValue(json_object, SettingNames::CONSIDERED_PROBE_TARGETS,
           considered_probe_targets_object);

  if (json_object.contains(SettingNames::CONSIDERED_PROBE_TARGETS)) {
    if (considered_probe_targets_object.contains("action")) {
      SetValue(considered_probe_targets_object, "action", action);
      if (action == "include") {
        are_probe_targets_included = true;
      } else if (action == "exclude") {
        are_probe_targets_included = false;
      } else {
        throw std::runtime_error(error_context + "Invalid action!");
      }

      if (considered_probe_targets_object.contains("signals")) {
        SetValue(considered_probe_targets_object, "signals", regex_pattern);
      } else {
        throw std::runtime_error(error_context + "Signal regex not found!");
      }
    } else {
      throw std::runtime_error(error_context + "Action not found!");
    }
  }

  considered_probe_targets = std::regex(regex_pattern);
}

void Settings::ParseConsideredExtensions(
    const boost::json::object& json_object) {
  std::string action, regex_pattern = ".*";
  boost::json::object considered_extensions_object;
  std::string error_context =
      "Error while parsing the considered probe targets: ";
  are_extensions_included = true;

  SetValue(json_object, SettingNames::CONSIDERED_EXTENSIONS,
           considered_extensions_object);

  if (json_object.contains(SettingNames::CONSIDERED_EXTENSIONS)) {
    if (considered_extensions_object.contains("action")) {
      SetValue(considered_extensions_object, "action", action);
      if (action == "include") {
        are_extensions_included = true;
      } else if (action == "exclude") {
        are_extensions_included = false;
      } else {
        throw std::runtime_error(error_context + "Invalid action!");
      }

      if (considered_extensions_object.contains("signals")) {
        SetValue(considered_extensions_object, "signals", regex_pattern);
      } else {
        throw std::runtime_error(error_context + "Signal regex not found!");
      }
    } else {
      throw std::runtime_error(error_context + "Action not found!");
    }
  }

  considered_extensions = std::regex(regex_pattern);
}

void Settings::ParseExpectedOutputs(const boost::json::object& json_object) {
  boost::json::array expected_output_values;
  std::string error_context = "Error while parsing the expected output values:";
  std::string signal_value;
  SetValue(json_object, SettingNames::EXPECTED_OUTPUT, expected_output_values);
  uint64_t bit_width;
  bool success;

  if (json_object.contains(SettingNames::EXPECTED_OUTPUT)) {
    if (expected_output_values.size() != groups.size()) {
      throw std::invalid_argument(error_context +
                                  "Size mismatch. The number of expected "
                                  "output values must be equal to "
                                  "the number of groups!");
    } else {
      for (auto& value : expected_output_values) {
        if (!value.is_string()) {
          throw std::invalid_argument(
              error_context +
              "Invalid format. The expected output values must be strings!");
        } else {
          signal_value = value.as_string().c_str();
          VerilogBitstringGrammar grammar(signal_value);
          expected_outputs.push_back(grammar.Parse(signal_value));
        }
      }
    }

    bit_width = expected_outputs[0].size();
    success = std::all_of(expected_outputs.begin(), expected_outputs.end(),
                          [bit_width](const std::vector<TriStateBit>& lhs) {
                            return lhs.size() == bit_width;
                          });

    if (!success) {
      throw std::invalid_argument(
          error_context +
          "Size mismatch. Every expected output must have the same bitwidth!");
    }
  }
}

void Settings::ParseHardwareTarget(const boost::json::object& json_object) {
  hardware_target = true;
  SetValue(json_object, SettingNames::HARDWARE_TARGET, hardware_target);
}

void Settings::ParseGroups(const boost::json::object& json_object) {
  boost::json::array groups_json;
  std::string error_context = "Error while parsing the group values: ";
  std::string input_value;
  uint64_t element_size = std::ceil(std::log2l(input_base)) * input_exponent;

  if (json_object.contains(SettingNames::GROUPS)) {
    SetValue(json_object, SettingNames::GROUPS, groups_json);
  } else {
    throw std::runtime_error(error_context + "The group values are missing.");
  }

  for (auto& group : groups_json) {
    if (!group.is_string()) {
      throw std::invalid_argument(error_context +
                                  "The group values must be given as strings.");
    } else {
      input_value = group.as_string().c_str();
      VerilogBitstringGrammar grammar(input_value);
      groups.push_back(grammar.Parse(input_value));

      if (groups.back().size() % element_size) {
        throw std::invalid_argument(
            error_context +
            "The group values must be a multiple of the element size defined "
            "with the input finite field.");
      }
    }
  }
}

void Settings::ParseLocationOfCipher(const boost::json::object& json_object) {
  location_of_cipher = "";
  if (json_object.contains(SettingNames::LOCATION_OF_CIPHER)) {
    SetValue(json_object, SettingNames::LOCATION_OF_CIPHER, location_of_cipher);
  } else {
    if (!hardware_target) {
      throw std::runtime_error(
          "Error while parsing the location of the cipher: The location of "
          "the "
          "cipher is missing.");
    }
  }
}

void Settings::ParseNumberOfClockCycles(
    const boost::json::object& json_object) {
  if (json_object.contains(SettingNames::NUMBER_OF_CLOCK_CYCLES)) {
    SetValue(json_object, SettingNames::NUMBER_OF_CLOCK_CYCLES,
             number_of_clock_cycles);
  } else {
    throw std::runtime_error(
        "Error while parsing the number of clock cycles: The number of "
        "clock cycles is missing.");
  }
}

void Settings::ParseMaximumMultivariateDistance(
    const boost::json::object& json_object) {
  maximum_multivariate_distance = 10;
  SetValue(json_object, SettingNames::MAXIMUM_MULTIVARIATE_DISTANCE,
           maximum_multivariate_distance);
}

void Settings::ParseMinimizeProbingSets(
    const boost::json::object& json_object) {
  minimize_probing_sets = false;
  SetValue(json_object, SettingNames::MINIMIZE_PROBING_SETS,
           minimize_probing_sets);
}

void Settings::ParseMultivariateTest(const boost::json::object& json_object) {
  std::string multivariate_test_string;
  multivariate_test = Analysis::univariate;
  SetValue(json_object, SettingNames::MULTIVARIATE_TEST,
           multivariate_test_string);

  if (multivariate_test_string == "multivariate") {
    multivariate_test = Analysis::multivariate;
  } else if ((multivariate_test_string == "exclusive_multivariate") ||
             (multivariate_test_string == "exclusive")) {
    multivariate_test = Analysis::exclusive_multivariate;
  }
}

void Settings::ParseNumberOfEntriesInReport(
    const boost::json::object& json_object) {
  number_of_entries_in_report = 0;
  SetValue(json_object, SettingNames::NUMBER_OF_ENTRIES_IN_REPORT,
           number_of_entries_in_report);
}

void Settings::ParseNumberOfPipelineStages(
    const boost::json::object& json_object) {
  number_of_pipeline_stages = 3;
  SetValue(json_object, SettingNames::NUMBER_OF_PIPELINE_STAGES,
           number_of_pipeline_stages);

  if (number_of_pipeline_stages > 6) {
    throw std::out_of_range(
        "Error while parsing the number of pipeline stages: The maximum "
        "supported number of pipeline stages is 6 (for the Cortex M7). If "
        "you "
        "require a higher number of pipeline stages please contact the "
        "Jannik "
        "Zeitschner (jannik.zeitschner@rub.de).");
  }
}

void Settings::ParseNumberOfProbingSetsPerStep(
    const boost::json::object& json_object) {
  number_of_probing_sets_per_step = UINT64_MAX;
  SetValue(json_object, SettingNames::NUMBER_OF_PROBING_SETS_PER_STEP,
           number_of_probing_sets_per_step);
}

void Settings::ParseNumberOfSimulationsPerStep(
    const boost::json::object& json_object) {
  number_of_simulations_per_step = 64;
  SetValue(json_object, SettingNames::NUMBER_OF_SIMULATIONS_PER_STEP,
           number_of_simulations_per_step);

  if ((number_of_simulations_per_step % 64 != 0) ||
      (number_of_simulations_per_step < 1)) {
    throw std::invalid_argument(
        "Error while parsing the number of step simulations: The number of "
        "step simulations must be a multiple of 64 and greater than 0.");
  }
}

void Settings::ParseNumberOfSimulationsPerWrite(
    const boost::json::object& json_object) {
  number_of_simulations_per_write = 0;
  SetValue(json_object, SettingNames::NUMBER_OF_SIMULATIONS_PER_WRITE,
           number_of_simulations_per_write);

  if (number_of_simulations_per_write) {
    if ((number_of_simulations_per_write % number_of_simulations_per_step) !=
        0) {
      throw std::invalid_argument(
          "Error while parsing the number of write simulations: The number of "
          "write simulations must be a multiple of the number of step "
          "simulations.");
    }
  }
}

void Settings::ParseNumberOfSimulations(
    const boost::json::object& json_object) {
  number_of_simulations = 64;
  SetValue(json_object, SettingNames::NUMBER_OF_SIMULATIONS,
           number_of_simulations);

  if (number_of_simulations % number_of_simulations_per_step != 0) {
    throw std::invalid_argument(
        "Error while parsing the number of simulations: The number of "
        "simulations must be a multiple of the number of step simulations.");
  }
}

void Settings::ParseNumberOfThreads(const boost::json::object& json_object) {
  std::string number_of_threads_as_string = "1";
  CpuCoreSelector cpu_core_selector;
  SetValue(json_object, SettingNames::NUMBER_OF_THREADS,
           number_of_threads_as_string);
  number_of_threads =
      cpu_core_selector.GetOptimalNumberOfThreads(number_of_threads_as_string);
}

void Settings::ParseOrderOfTest(const boost::json::object& json_object) {
  order_of_test = 1;
  SetValue(json_object, SettingNames::ORDER_OF_TEST, order_of_test);
}

void Settings::ParseOutputShares(const boost::json::object& json_object) {
  boost::json::array outputs;
  std::string error_context = "Error while parsing the output shares: ";
  std::string signal_name;
  SignalNameGrammar grammar;
  uint64_t bit_width;
  bool success;

  SetValue(json_object, SettingNames::OUTPUT_SHARES, outputs);
  if (json_object.contains(SettingNames::OUTPUT_SHARES)) {
    for (auto& output : outputs) {
      if (!output.is_string()) {
        throw std::invalid_argument(
            error_context +
            "Invalid format. The output shares must be given as strings.");
      } else {
        signal_name = output.as_string().c_str();
        output_shares.push_back(grammar.Parse(signal_name));
      }
    }

    bit_width = output_shares[0].size();
    success = std::all_of(output_shares.begin(), output_shares.end(),
                          [bit_width](const std::vector<std::string>& lhs) {
                            return lhs.size() == bit_width;
                          });

    if (!success) {
      throw std::invalid_argument(
          error_context +
          "Size mismatch. Every output share must have the same bitwidth!");
    }
  }
}

void Settings::ParseRemoveFullProbingSets(
    const boost::json::object& json_object) {
  remove_full_probing_sets = false;
  SetValue(json_object, SettingNames::REMOVE_FULL_PROBING_SETS,
           remove_full_probing_sets);
}

void Settings::ParseTestClockCycles(const boost::json::object& json_object) {
  test_clock_cycles.clear();
  std::string input,
      error_context = "Error while parsing the test clock cycles: ";
  boost::json::array ranges;
  IntegerRangeGrammar grammar(number_of_clock_cycles);
  std::pair<uint64_t, uint64_t> borders;
  uint64_t clock_cycle;
  SetValue(json_object, SettingNames::TEST_CLOCK_CYCLES, ranges);

  if (json_object.contains(SettingNames::TEST_CLOCK_CYCLES)) {
    for (auto& range : ranges) {
      if (!range.is_string()) {
        throw std::invalid_argument(
            error_context +
            "Every range of test clock cycles must be given as string.");
      }

      input = range.as_string().c_str();
      borders = grammar.Parse(input);
      for (clock_cycle = borders.first; clock_cycle <= borders.second;
           ++clock_cycle) {
        test_clock_cycles.push_back(clock_cycle);
      }
    }

    std::sort(test_clock_cycles.begin(), test_clock_cycles.end());
    test_clock_cycles.erase(
        std::unique(test_clock_cycles.begin(), test_clock_cycles.end()),
        test_clock_cycles.end());
  } else {
    for (uint64_t index = 1; index <= number_of_clock_cycles; ++index) {
      test_clock_cycles.push_back(index);
    }
  }
}

void Settings::ParseTransitionalLeakage(
    const boost::json::object& json_object) {
  transitional_leakage = false;
  SetValue(json_object, SettingNames::TRANSITIONAL_LEAKAGE,
           transitional_leakage);
}

void Settings::ParseRelaxedModel(const boost::json::object& json_object) {
  relaxed_model = false;
  SetValue(json_object, SettingNames::RELAXED_MODEL, relaxed_model);
}

void Settings::ParseWaveformSimulation(const boost::json::object& json_object) {
  waveform_simulation = false;
  SetValue(json_object, SettingNames::WAVEFORM_SIMULATION, waveform_simulation);
}

void Settings::ParseInputSequence(const boost::json::object& json_object) {
  boost::json::array input_sequence_json, signals_json;
  boost::json::object clock_cycle_obj, signal_obj;
  InputAssignmentGrammar value_grammar;
  SignalNameGrammar name_grammar;
  InputAssignment input_assignment;
  std::string signal_name, type_name, value_name;
  std::string error_context = "Error while parsing the input sequence: ";
  SetValue(json_object, SettingNames::INPUT_SEQUENCE, input_sequence_json);
  std::vector<InputAssignment> inputs_per_cycle;

  for (auto& clock_cycle : input_sequence_json) {
    if (!clock_cycle.is_object()) {
      throw std::runtime_error(
          error_context + "Invalid type. Every element must be a json object!");
    } else {
      clock_cycle_obj = clock_cycle.as_object();

      if (!clock_cycle_obj.contains("signals")) {
        throw std::runtime_error(
            error_context +
            "Missing type. Every object must have key signals!");
      }

      if (!clock_cycle_obj.at("signals").is_array()) {
        throw std::runtime_error(
            error_context +
            "Invalid type. Every element must be a json array!");
      }

      signals_json = clock_cycle_obj.at("signals").as_array();

      for (auto& signal : signals_json) {
        if (!signal.is_object()) {
          throw std::runtime_error(
              error_context +
              "Invalid type. Every element must be a json object!");
        }

        signal_obj = signal.as_object();

        if (!signal_obj.contains("name") || !signal_obj.contains("value")) {
          throw std::runtime_error(
              error_context +
              "Missing type. Every object must have name and value!");
        }

        if (!signal_obj.at("name").is_string() ||
            !signal_obj.at("value").is_string()) {
          throw std::runtime_error(
              error_context +
              "Invalid type. Both name and value must be given as strings!");
        }

        value_name = signal_obj.at("value").as_string().c_str();
        input_assignment = value_grammar.Parse(value_name);
        signal_name = signal_obj.at("name").as_string().c_str();
        input_assignment.signal_names_ = name_grammar.Parse(signal_name);

        if (input_assignment.signal_names_.size() !=
                input_assignment.group_indices_.size() &&
            input_assignment.signal_names_.size() !=
                input_assignment.signal_values_.size()) {
          throw std::runtime_error(
              error_context +
              "Size mismatch! The number of signal names in \"" + signal_name +
              "\" must be equal to the number of signal values in \"" +
              value_name + "\"!");
        }

        if (!signal_obj.contains("type")) {
          input_assignment.type_ = ValueType::standard;
        } else {
          if (!signal_obj.at("type").is_string()) {
            throw std::runtime_error(
                error_context +
                "Invalid type. The signal type must be given as strings!");
          }

          type_name = signal_obj.at("type").as_string().c_str();

          if (type_name == "var") {
            input_assignment.type_ = ValueType::variable;
          } else if (type_name == "arr") {
            input_assignment.type_ = ValueType::arary;
          } else {
            throw std::runtime_error(error_context +
                                     "Invalid type. The signal type must "
                                     "either be \"var\" or \"arr\"!");
          }
        }

        inputs_per_cycle.push_back(input_assignment);
      }

      input_sequence.push_back(inputs_per_cycle);
      inputs_per_cycle.clear();

      if (clock_cycle_obj.contains("hold_for_cycles")) {
        if (!clock_cycle_obj.at("hold_for_cycles").is_int64()) {
          throw std::runtime_error(error_context +
                                   "Invalid type. \"hold_for_cycles\" is not "
                                   "given as unsigned integer!");
        }

        for (int64_t index = 1;
             index < clock_cycle_obj.at("hold_for_cycles").as_int64();
             ++index) {
          input_sequence.push_back(inputs_per_cycle);
        }
      }
    }
  }
}

Settings::Settings(const std::string& config_file_path) {
  boost::json::object json_object = ParseJsonFile(config_file_path);
  std::cout << "Successfully parsed the settings file at \"" << config_file_path
            << "\"." << std::endl;
  ParseInputFiniteField(json_object);
  ParseOutputFiniteField(json_object);
  ParseNumberOfThreads(json_object);
  ParseHardwareTarget(json_object);
  ParseNumberOfClockCycles(json_object);
  ParseGroups(json_object);
  ParseAlwaysRandomInputs(json_object);
  ParseClockSignalName(json_object);
  ParseCompactDistributions(json_object);
  ParseCompilerFlags(json_object);
  ParseEffectSize(json_object);
  ParseEndCondition(json_object);
  ParseEndWaitCycles(json_object);
  ParseConsideredProbeTargets(json_object);
  ParseConsideredExtensions(json_object);
  ParseExpectedOutputs(json_object);
  ParseLocationOfCipher(json_object);
  ParseMaximumMultivariateDistance(json_object);
  ParseMinimizeProbingSets(json_object);
  ParseMultivariateTest(json_object);
  ParseNumberOfEntriesInReport(json_object);
  ParseNumberOfPipelineStages(json_object);
  ParseNumberOfProbingSetsPerStep(json_object);
  ParseNumberOfSimulationsPerStep(json_object);
  ParseNumberOfSimulationsPerWrite(json_object);
  ParseNumberOfSimulations(json_object);
  ParseOrderOfTest(json_object);
  ParseOutputShares(json_object);  // nicht in sw
  ParseRelaxedModel(json_object);
  ParseRemoveFullProbingSets(json_object);
  ParseTransitionalLeakage(json_object);
  ParseTestClockCycles(json_object);
  ParseWaveformSimulation(json_object);
  ParseInputSequence(json_object);
  ParseAlphaThreshold(json_object);
  ParseBetaThreshold(json_object);
}

const std::string SettingNames::CLOCK_SIGNAL_NAME = "clock_signal_name";
const std::string SettingNames::COMPACT_DISTRIBUTIONS = "compact_distributions";
const std::string SettingNames::EFFECT_SIZE = "effect_size";
const std::string SettingNames::END_CONDITION = "end_condition";
const std::string SettingNames::END_WAIT_CYCLES = "end_wait_cycles";
const std::string SettingNames::EXPECTED_OUTPUT = "expected_output";
const std::string SettingNames::NUMBER_OF_CLOCK_CYCLES =
    "number_of_clock_cycles";
const std::string SettingNames::MAXIMUM_MULTIVARIATE_DISTANCE =
    "max_distance_multivariate";
const std::string SettingNames::NUMBER_OF_THREADS = "max_number_of_threads";
const std::string SettingNames::MINIMIZE_PROBING_SETS = "minimize_probing_sets";
const std::string SettingNames::MULTIVARIATE_TEST = "multivariate_test";
const std::string SettingNames::ALWAYS_RANDOM_INPUTS = "always_random_inputs";
const std::string SettingNames::NUMBER_OF_ENTRIES_IN_REPORT =
    "number_of_entries_in_report";
const std::string SettingNames::HARDWARE_TARGET = "hardware_target";
const std::string SettingNames::GROUPS = "groups";
const std::string SettingNames::INPUT_SEQUENCE = "input_sequence";
const std::string SettingNames::NUMBER_OF_INITIAL_INPUTS =
    "number_of_initial_inputs";
const std::string SettingNames::OUTPUT_SHARES = "output_shares";
const std::string SettingNames::NUMBER_OF_PROBING_SETS_PER_STEP =
    "number_of_probing_sets_per_step";
const std::string SettingNames::NUMBER_OF_SIMULATIONS = "number_of_simulations";
const std::string SettingNames::NUMBER_OF_SIMULATIONS_PER_STEP =
    "number_of_simulations_per_step";
const std::string SettingNames::NUMBER_OF_SIMULATIONS_PER_WRITE =
    "number_of_simulations_per_write";
const std::string SettingNames::TEST_CLOCK_CYCLES = "test_clock_cycles";
const std::string SettingNames::ORDER_OF_TEST = "order_of_test";
const std::string SettingNames::CONSIDERED_PROBE_TARGETS =
    "considered_probe_targets";
const std::string SettingNames::CONSIDERED_EXTENSIONS = "considered_extensions";
const std::string SettingNames::REMOVE_FULL_PROBING_SETS =
    "remove_full_probing_sets";
const std::string SettingNames::TRANSITIONAL_LEAKAGE = "transitional_leakage";
const std::string SettingNames::WAVEFORM_SIMULATION = "waveform_simulation";
const std::string SettingNames::COMPILER_FLAGS = "compiler_flags";
const std::string SettingNames::LOCATION_OF_CIPHER = "location_of_cipher";
const std::string SettingNames::NUMBER_OF_PIPELINE_STAGES =
    "number_of_pipeline_stages";
const std::string SettingNames::RELAXED_MODEL = "relaxed_model";
const std::string SettingNames::ALPHA_THRESHOLD = "alpha_threshold";
const std::string SettingNames::BETA_THRESHOLD = "beta_threshold";
const std::string SettingNames::INPUT_FINITE_FIELD = "input_finite_field";
const std::string SettingNames::OUTPUT_FINITE_FIELD = "output_finite_field";
